@@ -3,9 +3,11 @@ from mpi4py import MPI as _MPI
 
 from jax import abstract_arrays, core
 from jax.core import Primitive, Tracer, Token
-from jax.interpreters import ad, xla, batching
+from jax.interpreters import ad, batching
 from jax.lax import create_token
-from jax.lib import xla_client
+
+from jax.interpreters import mlir
+import jaxlib.mlir.ir as ir
 
 from ..utils import (
     HashableMPIType,
@@ -14,13 +16,15 @@ from ..utils import (
     to_mpi_handle,
     unpack_hashable,
     wrap_as_hashable,
-    xla_constant_intc,
-    xla_constant_uintptr,
+    as_mhlo_constant,
+    get_default_layouts,
+    effect,
 )
+from ..jax_compat import hlo_custom_call
 from ..decorators import translation_rule_cpu, translation_rule_gpu
 from ..validation import enforce_types
 from ..comm import get_default_comm
-from ..jax_compat import register_abstract_eval
+
 
 # The Jax primitive
 mpi_allreduce_p = Primitive("allreduce_mpi")  # Create the primitive
@@ -70,43 +74,50 @@ def allreduce(x, op, *, comm=None, token=None):
 # transpose is a boolean flag that signals whever this is the forward pass
 # performing the MPI reduction, or the transposed pass, which is trivial
 @translation_rule_cpu
-def mpi_allreduce_xla_encode_cpu(c, x, token, op, comm, transpose):
+def mpi_allreduce_xla_encode_cpu(ctx, x, token, op, comm, transpose):
     op = unpack_hashable(op)
     comm = unpack_hashable(comm)
 
     if transpose:
         assert op == _MPI.SUM
-        return xla_client.ops.Tuple(c, [x, token])
+        return [x, token]
 
-    x_shape = c.GetShape(x)
-    dtype = x_shape.element_type()
-    dims = x_shape.dimensions()
+    x_aval, *_ = ctx.avals_in
+    x_nptype = x_aval.dtype
+
+    x_type = ir.RankedTensorType(x.type)
+    dtype = x_type.element_type
+    dims = x_type.shape
 
     # compute total number of elements in array
     nitems = _np.prod(dims, dtype=int)
 
-    sh = xla_client.Shape.tuple_shape(
-        [xla_client.Shape.array_shape(dtype, dims), xla_client.Shape.token_shape()]
+    out_types = [
+        ir.RankedTensorType.get(dims, dtype),
+        *mlir.token_type(),
+    ]
+
+    operands = (
+        as_mhlo_constant(nitems, _np.intc),
+        x,
+        as_mhlo_constant(to_mpi_handle(op), _np.uintp),
+        as_mhlo_constant(to_mpi_handle(comm), _np.uintp),
+        as_mhlo_constant(to_dtype_handle(x_nptype), _np.uintp),
+        token,
     )
 
-    return xla_client.ops.CustomCall(
-        c,
+    return hlo_custom_call(
         b"mpi_allreduce",
-        operands=(
-            xla_constant_intc(c, nitems),
-            x,
-            xla_constant_uintptr(c, to_mpi_handle(op)),
-            xla_constant_uintptr(c, to_mpi_handle(comm)),
-            xla_constant_uintptr(c, to_dtype_handle(dtype)),
-            token,
-        ),
-        shape=sh,
+        out_types=out_types,
+        operands=operands,
+        operand_layouts=get_default_layouts(operands),
+        result_layouts=get_default_layouts(out_types),
         has_side_effect=True,
     )
 
 
 @translation_rule_gpu
-def mpi_allreduce_xla_encode_gpu(c, x, token, op, comm, transpose):
+def mpi_allreduce_xla_encode_gpu(ctx, x, token, op, comm, transpose):
     from ..xla_bridge.mpi_xla_bridge_gpu import build_allreduce_descriptor
 
     op = unpack_hashable(op)
@@ -114,36 +125,43 @@ def mpi_allreduce_xla_encode_gpu(c, x, token, op, comm, transpose):
 
     if transpose:
         assert op == _MPI.SUM
-        return xla_client.ops.Tuple(c, [x, token])
+        return [x, token]
 
-    x_shape = c.GetShape(x)
-    dtype = x_shape.element_type()
-    dims = x_shape.dimensions()
+    x_aval, *_ = ctx.avals_in
+    x_nptype = x_aval.dtype
+
+    x_type = ir.RankedTensorType(x.type)
+    dtype = x_type.element_type
+    dims = x_type.shape
 
     # compute total number of elements in array
     nitems = _np.prod(dims, dtype=int)
 
-    sh = xla_client.Shape.tuple_shape(
-        [xla_client.Shape.array_shape(dtype, dims), xla_client.Shape.token_shape()]
+    out_types = [
+        ir.RankedTensorType.get(dims, dtype),
+        *mlir.token_type(),
+    ]
+
+    operands = (
+        x,
+        token,
     )
 
     descriptor = build_allreduce_descriptor(
         _np.intc(nitems),
         to_mpi_handle(op),
         to_mpi_handle(comm),
-        to_dtype_handle(dtype),
+        to_dtype_handle(x_nptype),
     )
 
-    return xla_client.ops.CustomCall(
-        c,
+    return hlo_custom_call(
         b"mpi_allreduce",
-        operands=(
-            x,
-            token,
-        ),
-        shape=sh,
-        opaque=descriptor,
+        out_types=out_types,
+        operands=operands,
+        operand_layouts=get_default_layouts(operands),
+        result_layouts=get_default_layouts(out_types),
         has_side_effect=True,
+        backend_config=descriptor,
     )
 
 
@@ -152,7 +170,7 @@ def mpi_allreduce_abstract_eval(xs, token, op, comm, transpose):
     return (
         abstract_arrays.ShapedArray(xs.shape, xs.dtype),
         core.abstract_token,
-    )
+    ), {effect}
 
 
 def mpi_allreduce_batch_eval(in_args, batch_axes, op, comm, transpose):
@@ -196,7 +214,7 @@ def mpi_allreduce_transpose_rule(tan_args, *x_args, op, comm, transpose):
 
 mpi_allreduce_p.multiple_results = True
 mpi_allreduce_p.def_impl(mpi_allreduce_impl)
-register_abstract_eval(mpi_allreduce_p, mpi_allreduce_abstract_eval)
+mpi_allreduce_p.def_effectful_abstract_eval(mpi_allreduce_abstract_eval)
 
 batching.primitive_batchers[mpi_allreduce_p] = mpi_allreduce_batch_eval
 
@@ -204,5 +222,5 @@ ad.primitive_jvps[mpi_allreduce_p] = mpi_allreduce_value_and_jvp
 ad.primitive_transposes[mpi_allreduce_p] = mpi_allreduce_transpose_rule
 
 # assign to the primitive the correct encoder
-xla.backend_specific_translations["cpu"][mpi_allreduce_p] = mpi_allreduce_xla_encode_cpu
-xla.backend_specific_translations["gpu"][mpi_allreduce_p] = mpi_allreduce_xla_encode_gpu
+mlir.register_lowering(mpi_allreduce_p, mpi_allreduce_xla_encode_cpu, platform="cpu")
+mlir.register_lowering(mpi_allreduce_p, mpi_allreduce_xla_encode_gpu, platform="cuda")
