@@ -3,9 +3,11 @@ from mpi4py import MPI as _MPI
 
 from jax import abstract_arrays, core
 from jax.core import Primitive, Tracer, Token
-from jax.interpreters import xla
 from jax.lax import create_token
-from jax.lib import xla_client
+
+from jax.interpreters import mlir
+import jaxlib.mlir.ir as ir
+from jaxlib.mhlo_helpers import custom_call
 
 from ..utils import (
     HashableMPIType,
@@ -14,13 +16,14 @@ from ..utils import (
     to_mpi_handle,
     unpack_hashable,
     wrap_as_hashable,
-    xla_constant_intc,
-    xla_constant_uintptr,
+    as_mhlo_constant,
+    get_default_layouts,
+    effect,
 )
 from ..decorators import translation_rule_cpu, translation_rule_gpu
 from ..validation import enforce_types
 from ..comm import get_default_comm
-from ..jax_compat import register_abstract_eval
+
 
 # The Jax primitive
 mpi_alltoall_p = Primitive("alltoall_mpi")  # Create the primitive
@@ -76,68 +79,77 @@ def alltoall(
 
 # This function compiles the operation
 @translation_rule_cpu
-def mpi_alltoall_xla_encode_cpu(c, x, token, comm):
+def mpi_alltoall_xla_encode_cpu(ctx, x, token, comm):
     comm = unpack_hashable(comm)
 
-    shape = c.GetShape(x)
-    dtype = shape.element_type()
-    dims = shape.dimensions()
+    x_aval, *_ = ctx.avals_in
+    x_nptype = x_aval.dtype
+
+    x_type = ir.RankedTensorType(x.type)
+    dtype = x_type.element_type
+    dims = x_type.shape
 
     # compute total number of elements in array
     size = comm.Get_size()
     assert dims[0] == size
     nitems_per_proc = _np.prod(dims[1:], dtype=int)
-    dtype_handle = to_dtype_handle(dtype)
+    dtype_handle = to_dtype_handle(x_nptype)
 
-    sh = xla_client.Shape.tuple_shape(
-        [
-            xla_client.Shape.array_shape(dtype, dims),
-            xla_client.Shape.token_shape(),
-        ]
-    )
+    out_types = [
+        ir.RankedTensorType.get(dims, dtype),
+        *mlir.token_type(),
+    ]
 
     operands = (
-        xla_constant_intc(c, nitems_per_proc),
+        as_mhlo_constant(nitems_per_proc, _np.intc),
         x,
-        xla_constant_uintptr(c, dtype_handle),
+        as_mhlo_constant(dtype_handle, _np.uintp),
         # we only support matching input and output arrays
-        xla_constant_intc(c, nitems_per_proc),
-        xla_constant_uintptr(c, dtype_handle),
+        as_mhlo_constant(nitems_per_proc, _np.intc),
+        as_mhlo_constant(dtype_handle, _np.uintp),
         #
-        xla_constant_uintptr(c, to_mpi_handle(comm)),
+        as_mhlo_constant(to_mpi_handle(comm), _np.uintp),
         token,
     )
 
-    return xla_client.ops.CustomCall(
-        c,
+    return custom_call(
         b"mpi_alltoall",
+        out_types=out_types,
         operands=operands,
-        shape=sh,
+        # force c order because first axis is special
+        operand_layouts=get_default_layouts(operands, order="c"),
+        result_layouts=get_default_layouts(out_types, order="c"),
         has_side_effect=True,
     )
 
 
 @translation_rule_gpu
-def mpi_alltoall_xla_encode_gpu(c, x, token, comm):
+def mpi_alltoall_xla_encode_gpu(ctx, x, token, comm):
     from ..xla_bridge.mpi_xla_bridge_gpu import build_alltoall_descriptor
 
     comm = unpack_hashable(comm)
 
-    shape = c.GetShape(x)
-    dtype = shape.element_type()
-    dims = shape.dimensions()
+    x_aval, *_ = ctx.avals_in
+    x_nptype = x_aval.dtype
 
-    # compute total number of elements in send array
+    x_type = ir.RankedTensorType(x.type)
+    dtype = x_type.element_type
+    dims = x_type.shape
+
+    # compute total number of elements in array
     size = comm.Get_size()
     assert dims[0] == size
     nitems_per_proc = _np.prod(dims[1:], dtype=int)
-    dtype_handle = to_dtype_handle(dtype)
+    dtype_handle = to_dtype_handle(x_nptype)
 
-    sh = xla_client.Shape.tuple_shape(
-        [
-            xla_client.Shape.array_shape(dtype, dims),
-            xla_client.Shape.token_shape(),
-        ]
+    out_types = [
+        ir.RankedTensorType.get(dims, dtype),
+        *mlir.token_type(),
+    ]
+
+    operands = (
+        x,
+        token,
     )
 
     descriptor = build_alltoall_descriptor(
@@ -150,16 +162,15 @@ def mpi_alltoall_xla_encode_gpu(c, x, token, comm):
         to_mpi_handle(comm),
     )
 
-    return xla_client.ops.CustomCall(
-        c,
+    return custom_call(
         b"mpi_alltoall",
-        operands=(
-            x,
-            token,
-        ),
-        shape=sh,
-        opaque=descriptor,
+        out_types=out_types,
+        operands=operands,
+        # force c order because first axis is special
+        operand_layouts=get_default_layouts(operands, order="c"),
+        result_layouts=get_default_layouts(out_types, order="c"),
         has_side_effect=True,
+        backend_config=descriptor,
     )
 
 
@@ -168,13 +179,13 @@ def mpi_alltoall_abstract_eval(xs, token, comm):
     return (
         abstract_arrays.ShapedArray(xs.shape, xs.dtype),
         core.abstract_token,
-    )
+    ), {effect}
 
 
 mpi_alltoall_p.multiple_results = True
 mpi_alltoall_p.def_impl(mpi_alltoall_impl)
-register_abstract_eval(mpi_alltoall_p, mpi_alltoall_abstract_eval)
+mpi_alltoall_p.def_effectful_abstract_eval(mpi_alltoall_abstract_eval)
 
 # assign to the primitive the correct encoder
-xla.backend_specific_translations["cpu"][mpi_alltoall_p] = mpi_alltoall_xla_encode_cpu
-xla.backend_specific_translations["gpu"][mpi_alltoall_p] = mpi_alltoall_xla_encode_gpu
+mlir.register_lowering(mpi_alltoall_p, mpi_alltoall_xla_encode_cpu, platform="cpu")
+mlir.register_lowering(mpi_alltoall_p, mpi_alltoall_xla_encode_gpu, platform="cuda")
