@@ -3,10 +3,11 @@ from mpi4py import MPI as _MPI
 
 from jax import abstract_arrays, core
 from jax.core import Primitive, Tracer, Token
-from jax.interpreters import xla
-from jax.lib import xla_client
 from jax.lax import create_token
 
+from jax.interpreters import mlir
+import jaxlib.mlir.ir as ir
+
 from ..utils import (
     HashableMPIType,
     default_primitive_impl,
@@ -14,13 +15,15 @@ from ..utils import (
     to_mpi_handle,
     unpack_hashable,
     wrap_as_hashable,
-    xla_constant_intc,
-    xla_constant_uintptr,
+    as_mhlo_constant,
+    get_default_layouts,
+    effect,
 )
+from ..jax_compat import hlo_custom_call
 from ..decorators import translation_rule_cpu, translation_rule_gpu
 from ..validation import enforce_types
 from ..comm import get_default_comm
-from ..jax_compat import register_abstract_eval
+
 
 # The Jax primitive
 mpi_scan_p = Primitive("scan_mpi")  # Create the primitive
@@ -63,57 +66,73 @@ def scan(x, op, *, comm=None, token=None):
 
 # This function compiles the operation
 @translation_rule_cpu
-def mpi_scan_xla_encode_cpu(c, x, token, op, comm):
+def mpi_scan_xla_encode_cpu(ctx, x, token, op, comm):
     op = unpack_hashable(op)
     comm = unpack_hashable(comm)
 
-    x_shape = c.GetShape(x)
-    dtype = x_shape.element_type()
-    dims = x_shape.dimensions()
+    x_aval, *_ = ctx.avals_in
+    x_nptype = x_aval.dtype
+
+    x_type = ir.RankedTensorType(x.type)
+    dtype = x_type.element_type
+    dims = x_type.shape
 
     # compute total number of elements in array
     nitems = _np.prod(dims, dtype=int)
 
-    dtype_handle = to_dtype_handle(dtype)
+    dtype_handle = to_dtype_handle(x_nptype)
 
-    sh = xla_client.Shape.tuple_shape(
-        [xla_client.Shape.array_shape(dtype, dims), xla_client.Shape.token_shape()]
+    out_types = [
+        ir.RankedTensorType.get(dims, dtype),
+        *mlir.token_type(),
+    ]
+
+    operands = (
+        as_mhlo_constant(nitems, _np.intc),
+        x,
+        as_mhlo_constant(to_mpi_handle(op), _np.uintp),
+        as_mhlo_constant(to_mpi_handle(comm), _np.uintp),
+        as_mhlo_constant(dtype_handle, _np.uintp),
+        token,
     )
 
-    return xla_client.ops.CustomCall(
-        c,
+    return hlo_custom_call(
         b"mpi_scan",
-        operands=(
-            xla_constant_intc(c, nitems),
-            x,
-            xla_client.ops.Constant(c, to_mpi_handle(op)),
-            xla_constant_uintptr(c, to_mpi_handle(comm)),
-            xla_constant_uintptr(c, dtype_handle),
-            token,
-        ),
-        shape=sh,
+        out_types=out_types,
+        operands=operands,
+        operand_layouts=get_default_layouts(operands),
+        result_layouts=get_default_layouts(out_types),
         has_side_effect=True,
     )
 
 
 @translation_rule_gpu
-def mpi_scan_xla_encode_gpu(c, x, token, op, comm):
+def mpi_scan_xla_encode_gpu(ctx, x, token, op, comm):
     from ..xla_bridge.mpi_xla_bridge_gpu import build_scan_descriptor
 
     op = unpack_hashable(op)
     comm = unpack_hashable(comm)
 
-    x_shape = c.GetShape(x)
-    dtype = x_shape.element_type()
-    dims = x_shape.dimensions()
+    x_aval, *_ = ctx.avals_in
+    x_nptype = x_aval.dtype
+
+    x_type = ir.RankedTensorType(x.type)
+    dtype = x_type.element_type
+    dims = x_type.shape
 
     # compute total number of elements in array
     nitems = _np.prod(dims, dtype=int)
 
-    dtype_handle = to_dtype_handle(dtype)
+    dtype_handle = to_dtype_handle(x_nptype)
 
-    sh = xla_client.Shape.tuple_shape(
-        [xla_client.Shape.array_shape(dtype, dims), xla_client.Shape.token_shape()]
+    out_types = [
+        ir.RankedTensorType.get(dims, dtype),
+        *mlir.token_type(),
+    ]
+
+    operands = (
+        x,
+        token,
     )
 
     descriptor = build_scan_descriptor(
@@ -123,16 +142,14 @@ def mpi_scan_xla_encode_gpu(c, x, token, op, comm):
         dtype_handle,
     )
 
-    return xla_client.ops.CustomCall(
-        c,
+    return hlo_custom_call(
         b"mpi_scan",
-        operands=(
-            x,
-            token,
-        ),
-        shape=sh,
-        opaque=descriptor,
+        out_types=out_types,
+        operands=operands,
+        operand_layouts=get_default_layouts(operands),
+        result_layouts=get_default_layouts(out_types),
         has_side_effect=True,
+        backend_config=descriptor,
     )
 
 
@@ -141,13 +158,13 @@ def mpi_scan_abstract_eval(xs, token, op, comm):
     return (
         abstract_arrays.ShapedArray(xs.shape, xs.dtype),
         core.abstract_token,
-    )
+    ), {effect}
 
 
 mpi_scan_p.multiple_results = True
 mpi_scan_p.def_impl(mpi_scan_impl)
-register_abstract_eval(mpi_scan_p, mpi_scan_abstract_eval)
+mpi_scan_p.def_effectful_abstract_eval(mpi_scan_abstract_eval)
 
 # assign to the primitive the correct encoder
-xla.backend_specific_translations["cpu"][mpi_scan_p] = mpi_scan_xla_encode_cpu
-xla.backend_specific_translations["gpu"][mpi_scan_p] = mpi_scan_xla_encode_gpu
+mlir.register_lowering(mpi_scan_p, mpi_scan_xla_encode_cpu, platform="cpu")
+mlir.register_lowering(mpi_scan_p, mpi_scan_xla_encode_gpu, platform="cuda")
