@@ -3,9 +3,11 @@ from mpi4py import MPI as _MPI
 
 from jax import core
 from jax.core import Primitive, Tracer, Token
-from jax.interpreters import xla
 from jax.lax import create_token
-from jax.lib import xla_client
+
+from jax.interpreters import mlir
+import jaxlib.mlir.ir as ir
+from jaxlib.mhlo_helpers import custom_call
 
 from ..utils import (
     HashableMPIType,
@@ -14,13 +16,14 @@ from ..utils import (
     to_mpi_handle,
     unpack_hashable,
     wrap_as_hashable,
-    xla_constant_intc,
-    xla_constant_uintptr,
+    as_mhlo_constant,
+    get_default_layouts,
+    effect,
 )
 from ..decorators import translation_rule_cpu, translation_rule_gpu
 from ..validation import enforce_types
 from ..comm import get_default_comm
-from ..jax_compat import register_abstract_eval
+
 
 # The Jax primitive
 mpi_send_p = Primitive("send_mpi")  # Create the primitive
@@ -62,55 +65,66 @@ def send(x, dest, *, tag=0, comm=None, token=None):
 
 # This function compiles the operation
 @translation_rule_cpu
-def mpi_send_xla_encode_cpu(c, x, token, dest, tag, comm):
+def mpi_send_xla_encode_cpu(ctx, x, token, dest, tag, comm):
     comm = unpack_hashable(comm)
 
-    x_shape = c.GetShape(x)
-    dtype = x_shape.element_type()
-    dims = x_shape.dimensions()
+    x_aval, *_ = ctx.avals_in
+    x_nptype = x_aval.dtype
+
+    x_type = ir.RankedTensorType(x.type)
+    dims = x_type.shape
 
     # compute total number of elements in array
     nitems = _np.prod(dims, dtype=int)
-    dtype_handle = to_dtype_handle(dtype)
+    dtype_handle = to_dtype_handle(x_nptype)
 
-    # ensure void** out type
-    sh = xla_client.Shape.tuple_shape([xla_client.Shape.token_shape()])
+    out_types = mlir.token_type()
 
-    out = xla_client.ops.CustomCall(
-        c,
-        b"mpi_send",
-        operands=(
-            xla_constant_intc(c, nitems),
-            x,
-            xla_client.ops.Constant(c, _np.intc(dest)),
-            xla_constant_intc(c, tag),
-            xla_constant_uintptr(c, to_mpi_handle(comm)),
-            xla_constant_uintptr(c, dtype_handle),
-            token,
-        ),
-        shape=sh,
-        has_side_effect=True,
+    operands = (
+        as_mhlo_constant(nitems, _np.intc),
+        x,
+        as_mhlo_constant(dest, _np.intc),
+        as_mhlo_constant(tag, _np.intc),
+        as_mhlo_constant(to_mpi_handle(comm), _np.uintp),
+        as_mhlo_constant(dtype_handle, _np.uintp),
+        token,
     )
 
-    return xla_client.ops.GetTupleElement(out, 0)
+    # JAX insists on outputs being iterable
+    return [
+        custom_call(
+            b"mpi_send",
+            out_types=out_types,
+            operands=operands,
+            operand_layouts=get_default_layouts(operands),
+            result_layouts=get_default_layouts(out_types),
+            has_side_effect=True,
+        )
+    ]
 
 
 @translation_rule_gpu
-def mpi_send_xla_encode_gpu(c, x, token, dest, tag, comm):
+def mpi_send_xla_encode_gpu(ctx, x, token, dest, tag, comm):
     from ..xla_bridge.mpi_xla_bridge_gpu import build_send_descriptor
 
     comm = unpack_hashable(comm)
 
-    x_shape = c.GetShape(x)
-    dtype = x_shape.element_type()
-    dims = x_shape.dimensions()
+    x_aval, *_ = ctx.avals_in
+    x_nptype = x_aval.dtype
+
+    x_type = ir.RankedTensorType(x.type)
+    dims = x_type.shape
 
     # compute total number of elements in array
     nitems = _np.prod(dims, dtype=int)
-    dtype_handle = to_dtype_handle(dtype)
+    dtype_handle = to_dtype_handle(x_nptype)
 
-    # ensure void** out type
-    sh = xla_client.Shape.tuple_shape([xla_client.Shape.token_shape()])
+    out_types = mlir.token_type()
+
+    operands = (
+        x,
+        token,
+    )
 
     descriptor = build_send_descriptor(
         nitems,
@@ -120,29 +134,28 @@ def mpi_send_xla_encode_gpu(c, x, token, dest, tag, comm):
         dtype_handle,
     )
 
-    out = xla_client.ops.CustomCall(
-        c,
-        b"mpi_send",
-        operands=(
-            x,
-            token,
-        ),
-        shape=sh,
-        opaque=descriptor,
-        has_side_effect=True,
-    )
-
-    return xla_client.ops.GetTupleElement(out, 0)
+    # JAX insists on outputs being iterable
+    return [
+        custom_call(
+            b"mpi_send",
+            out_types=out_types,
+            operands=operands,
+            operand_layouts=get_default_layouts(operands),
+            result_layouts=get_default_layouts(out_types),
+            has_side_effect=True,
+            backend_config=descriptor,
+        )
+    ]
 
 
 # This function evaluates only the shapes during AST construction
 def mpi_send_abstract_eval(xs, token, dest, tag, comm):
-    return core.abstract_token
+    return core.abstract_token, {effect}
 
 
 mpi_send_p.def_impl(mpi_send_impl)
-register_abstract_eval(mpi_send_p, mpi_send_abstract_eval)
+mpi_send_p.def_effectful_abstract_eval(mpi_send_abstract_eval)
 
 # assign to the primitive the correct encoder
-xla.backend_specific_translations["cpu"][mpi_send_p] = mpi_send_xla_encode_cpu
-xla.backend_specific_translations["gpu"][mpi_send_p] = mpi_send_xla_encode_gpu
+mlir.register_lowering(mpi_send_p, mpi_send_xla_encode_cpu, platform="cpu")
+mlir.register_lowering(mpi_send_p, mpi_send_xla_encode_gpu, platform="cuda")
